using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Luski.net.Enums;

namespace Luski.net;

public class API
{
    public MainServer MainServer { get; internal set; }
    internal List<PublicServer> InternalServers { get; } = new();
    public IReadOnlyList<PublicServer> LoadedServers => InternalServers.AsReadOnly();
    
    public async Task<PublicServer> GetPublicServer(string Domain, string Version = "v1", bool Secure = true)
    {
        DateTime dt = DateTime.UtcNow;
        Console.WriteLine("Connecting to public server '{0}' using API {1}.", Domain, Version);
        PublicServer s;
        try
        {
            IEnumerable<PublicServer> isl = InternalServers.Where(a => (a.Domain == Domain && a.ApiVersion == Version));
            if (isl.Any()) return isl.First();
            s = new(Domain, Version, Secure)
            {
                ServerType = ServerType.Public
            };
        }
        catch (Exception e)
        {
            Console.WriteLine("Failed to connect to public server '{0}' using API {1}.", Domain, Version);
            throw;
        }
        Console.WriteLine("Connected to public server '{0}' using API {1} in {4}.\nServer Name: {2}\nServer Description: {3}", Domain, Version, s.Name, s.Description, DateTime.UtcNow.Subtract(dt).ToString("g"));
        string? f = s.Storage.GetStorageDirectory(StorageDirectory.StorageInfo) + "token";
        if (File.Exists(f))
        {
            f = File.ReadAllText(f);
        }
        else f = null;

        try
        {
            if (f is not null)
            {
                bool b  = await s.LoginViaToken(f);
                if (b) Console.WriteLine("Auto Login Successful");
                else
                {
                    s.Token = null;
                    s.Error = null;
                }
            }
        }
        catch
        {
        }

        InternalServers.Add(s);
        return s;
    }
    
    public MainServer GetMainServer(string Domain, string Version = "v1")
    {
        DateTime dt = DateTime.UtcNow;
        Console.WriteLine("Conecting to main server '{0}' using API {1}.", Domain, Version);
        MainServer = new(Domain, Version)
        {
            ServerType = ServerType.Main
        };
        Console.WriteLine("Connected to main server '{0}' using API {1} in {2}.", Domain, Version, DateTime.UtcNow.Subtract(dt).ToString("g"));
        return MainServer;
    }
}